#pragma once

#ifndef MC_LIB_INCLUDED
    #error need to include mc lib first
#endif

#include <avr/pgmspace.h>

//------------------------------------------------------------------------
// DISPLAY CONSTANTS
//------------------------------------------------------------------------
struct LCD5110_conf {
    constexpr static uint8_t width        = 84;
    constexpr static uint8_t height       = 48;
    constexpr static uint8_t glyph_width  = 1;
    constexpr static uint8_t glyph_height = 8;
    constexpr static uint8_t rows         = height / glyph_height;
    constexpr static uint8_t cols         = width  / glyph_width;
};

//------------------------------------------------------------------------
struct LCD5110_commands {
    constexpr static mc::word_t powerdown            = 0x04;
    constexpr static mc::word_t entrymode            = 0x02;
    constexpr static mc::word_t extended_instruction = 0x01;
    constexpr static mc::word_t display_blank        = 0x00;
    constexpr static mc::word_t display_normal       = 0x04;
    constexpr static mc::word_t display_allon        = 0x01;
    constexpr static mc::word_t display_inverted     = 0x05;
    constexpr static mc::word_t function_set         = 0x20;
    constexpr static mc::word_t display_control      = 0x08;
    constexpr static mc::word_t set_y                = 0x40;
    constexpr static mc::word_t set_x                = 0x80;
    constexpr static mc::word_t set_temp             = 0x04;
    constexpr static mc::word_t set_bias             = 0x10;
    constexpr static mc::word_t set_vop              = 0x80;
};

//------------------------------------------------------------------------
struct LCD5110_default_values {
    constexpr static mc::word_t bias     = 0x03;
    constexpr static mc::word_t temp     = 0x02;
    constexpr static mc::word_t contrast = 0x46;
};

//------------------------------------------------------------------------
// DISPLAY ACCESS
//------------------------------------------------------------------------
template<class SPI,
         const mc::pin::id_t& PinDC,
         const mc::pin::id_t& PinRST,
         const mc::pin::id_t& PinCS>
struct LCD5110 {

    //--------------------------------------------------------------------
    using conf           = LCD5110_conf;
    using cmd            = LCD5110_commands;
    using default_values = LCD5110_default_values;
    using glyph_t        = uint8_t;

    //--------------------------------------------------------------------
    LCD5110() {
        using namespace mc;

        mc::pin::configure<PinRST> (OUTPUT);
        mc::pin::configure<PinCS>  (OUTPUT);
        mc::pin::configure<PinDC>  (OUTPUT);

        // everything is active low
        pin::set<PinDC>(HIGH);
        pin::set<PinCS>(HIGH);
        util::blink<PinRST,LOW>(1,10);

        spi_.reset();

        // display init sequence
        send_command(cmd::function_set    | cmd::extended_instruction);
        send_command(cmd::set_vop         | default_values::contrast );
        send_command(cmd::set_temp        | default_values::temp     );
        send_command(cmd::set_bias        | default_values::bias     );
        send_command(cmd::function_set                               );
        send_command(cmd::display_control | cmd::display_normal      );
    }

    //--------------------------------------------------------------------
    void move(const uint8_t col, const uint8_t row) {
        send_command(cmd::set_x | row);
        send_command(cmd::set_y | col);
    }

    //--------------------------------------------------------------------
    void put(const glyph_t g) {
        send_data(g);
    }

    //--------------------------------------------------------------------
    void clear() {
        move(0,0);

        for (uint16_t i = 0; i < conf::rows * conf::cols; ++i) {
            put(0);
        }
    }

private:
    //--------------------------------------------------------------------
    inline void send_command(mc::word_t command) { send(command, false); }
    inline void send_data   (mc::word_t data   ) { send(data   , true ); }

    inline void send(mc::word_t byte, bool is_data) {
        using namespace mc;

        util::level_guard<PinCS> lgCS(LOW);
        pin::set<PinDC>(is_data ? HIGH : LOW);
        spi_(byte);
    }

    //--------------------------------------------------------------------
    mc::util::spi::context<mc::device::spi> spi_ctx_;
    mc::util::spi::master <mc::device::spi> spi_;
};

//------------------------------------------------------------------------
// TEXT
//------------------------------------------------------------------------

//------------------------------------------------------------------------
extern const mc::word_t LCD5110_default_font[] PROGMEM;

struct LCD5110_default_font_conf {
    constexpr static uint8_t glyph_width  = 6;
    constexpr static uint8_t glyph_height = 1;
    constexpr static uint8_t ascii_offset = 32;
    constexpr static const mc::word_t* font  = &(LCD5110_default_font[0]);
};

//------------------------------------------------------------------------
template<class LCD, class FontConfig = LCD5110_default_font_conf>
struct LCD5110_TypeWriter {
    using char_t = char;
    using conf   = FontConfig;

    LCD5110_TypeWriter(LCD& lcd): lcd_(lcd) {}

    void putchar(const char_t c) {
        static_assert(conf::glyph_height == 1, "needs glyph height of 1.");

        for (uint8_t i = 0; i < conf::glyph_width; ++i) {
            const auto base = (c - conf::ascii_offset) * conf::glyph_width;
            lcd_.put(pgm_read_byte(&(conf::font[base + i])));
        }
    }

private:
    LCD& lcd_;
};
