#pragma once

namespace mc { namespace util { namespace twi {

//------------------------------------------------------------------------------
namespace detail {

//------------------------------------------------------------------------------
template<class Twi>
struct base {

    //--------------------------------------------------------------------------
    base(bool& error_state, word_t& error_code)
    : error_state_(error_state),
      error_code_ (error_code)
    {}

    //--------------------------------------------------------------------------
    void expect(word_t s) {
        const auto& got = Twi::status();

        if (got != s) {
            error_state_ = true;
            error_code_  = got;
        }
    }

    //--------------------------------------------------------------------------
    operator bool() const {
        return (!error_state_);
    }

    //--------------------------------------------------------------------------
    inline void step() {
        Twi::step();
    }

    //--------------------------------------------------------------------------
    void wait() {
        while(! Twi::transmit_complete()) {}
    }

protected:
    //--------------------------------------------------------------------------
    bool&   error_state_;
    word_t& error_code_;
};

//------------------------------------------------------------------------------
} // namespace detail

//------------------------------------------------------------------------------
// MASTER
//------------------------------------------------------------------------------
template<class Twi>
struct master : public detail::base<Twi> {

    //--------------------------------------------------------------------------
    master(bool& error_state, word_t& error_code_)
    : detail::base<Twi>(error_state, error_code_)
    {
        Twi::start();
        this->wait();
        this->expect(TW_START);
    }

    ~master() {
        Twi::stop();
        while(! Twi::stop_complete());
    }

    //--------------------------------------------------------------------------
    master& rep_start() {
        Twi::start();
        this->wait();
        this->expect(TW_REP_START);

        need_rep_start_ = false;
        return *this;
    }

    //--------------------------------------------------------------------------
    struct write_mode : public detail::base<Twi> {
        write_mode(bool& error_state, word_t& error_code, word_t addr)
        : detail::base<Twi>(error_state, error_code)
        {
            Twi::template write_address<Twi::mode::write>(addr);
            this->step();
            this->wait();
            this->expect(TW_MT_SLA_ACK);
        }

        write_mode& operator()(word_t data) {
            Twi::write(data);
            this->step();
            this->wait();
            this->expect(TW_MT_DATA_ACK);
            return *this;
        }
    };

    inline write_mode write(word_t addr) {
        if (need_rep_start_) this->rep_start();
        need_rep_start_ = true;
        return {this->error_state_, this->error_code_, addr};
    }

    template<class Array>
    master& push(word_t addr, Array& arr) {
        auto w = this->write(addr);
        if (! w) return *this;

        for (auto i = 0; i < arr.size(); ++i) {
            if (! w(arr[i])) return *this;
        }

        return *this;
    }

    inline master& put(word_t addr, word_t data) {
        auto w = this->write(addr);
        if (! w)       return *this;
        if (! w(data)) return *this;
        return *this;
    }

    //--------------------------------------------------------------------------
    struct read_mode : public detail::base<Twi> {
        read_mode(bool& error_state, word_t& error_code, word_t addr)
        : detail::base<Twi>(error_state, error_code)
        {
            Twi::template write_address<Twi::mode::read>(addr);
            this->step();
            this->wait();
            this->expect(TW_MR_SLA_ACK);
        }

        read_mode& ack() {
            Twi::ack();
            this->wait();
            this->expect(TW_MR_DATA_ACK);
            return *this;
        }

        read_mode& nack() {
            Twi::nack();
            this->wait();
            this->expect(TW_MR_DATA_NACK);
            return *this;
        }

        inline word_t operator*() {
            return Twi::read();
        }
    };

    inline read_mode read(word_t addr) {
        if (need_rep_start_) this->rep_start();
        need_rep_start_ = true;
        return {this->error_state_, this->error_code_, addr};
    }

    template<class Array>
    master& fetch(word_t addr, Array& arr) {
        auto  r = this->read(addr);
        auto* p = arr.begin();

        if (! r) { return *this; }

        for (typename Array::count_t i = arr.size(); i > 0; --i) {
            if (i > 1) { if (!r.ack() ) return *this; }
            else       { if (!r.nack()) return *this; }
            *p++ = *r;
        }

        return *this;
    }

    inline master& get(word_t addr, word_t& buf) {
        auto r = this->read(addr);
        if (! r)        return *this;
        if (! r.nack()) return *this;
        buf = *r;
        return *this;
    }

//------------------------------------------------------------------------------
private:
    bool need_rep_start_ = false;
};

//------------------------------------------------------------------------------
// CONTEXT
//------------------------------------------------------------------------------
template<class Twi, typename Twi::freq_t Frequency>
struct context {

    //--------------------------------------------------------------------------
    context() {
        Twi::template set_frequency<Frequency>();
        Twi::enable();
    }

    ~context() {
        Twi::disable();
    }

    //--------------------------------------------------------------------------
    mc::util::twi::master<Twi> master() { return {error_state_, error_code_}; }

    //--------------------------------------------------------------------------
    operator bool()  const { return !error_state_; }
    word_t   error() const { return error_code_;   }

    void clear_error() {
        error_code_  = false;
        error_state_ = 0;
    }

private: 
    bool   error_state_ = false;
    word_t error_code_  = 0x0;
};

//------------------------------------------------------------------------------
}}} // namespace mc::util::twi
