#include "cpu/atmega328p.hpp"
#include "mc.hpp"
#include "external/display/lcd5110.hpp"

using namespace mc;

MC_USE_DYNAMIC_IRQS()
MC_ENABLE_DYNAMIC_IRQ(INT0_vect_num)
MC_ENABLE_DYNAMIC_IRQ(ADC_vect_num)

constexpr const auto* ledPin      = &pin::pc3;
constexpr const auto* btnPin      = &pin::pd2;
constexpr const auto* internalLed = &pin::pb5;
constexpr const auto* poti1Pin    = &pin::pc6;
constexpr const auto* poti2Pin    = &pin::pc5;
constexpr const auto* blPin       = &pin::pwm::oc2b;

using LCD = LCD5110<device::spi,
                    pin::pb1,
                    pin::pb0,
                    pin::pb2>;

static LCD                        lcd;
static LCD5110_TypeWriter<LCD>    tw(lcd);
static io::ostream<decltype(tw) > lout(tw);

static device::uart              uart_inst;
static io::ostream<device::uart> uout(uart_inst);
static io::istream<device::uart> uin(uart_inst);

void signalStartup() {
    util::blink<*internalLed>();
}

void buttonPressed() {
    util::blink<*ledPin>();
}

void setBrightness() {
    device::timer2::set_ocrb(util::adc::read<device::adc::r8>());
    device::irq::enable<device::irq::timer1_ovf>();
}

void displayADC() {
    // read adc and print to display and serial device
    const auto adc_val = util::adc::read();
    uout << "poti: " << adc_val << "\n";

    lcd.move(0,0); lout << "                       ";
    lcd.move(0,0); lout << "-->" << adc_val << "<--";

    // read poti2, then call setBrightness
    device::irq::attach<device::irq::adc>(setBrightness);
    util::adc::start_async<*poti2Pin, device::adc::r8>();
}

void startReadADC() {
    device::irq::disable<device::irq::timer1_ovf>();

    // read poti1, when finished call displayADC
    device::irq::attach<device::irq::adc>(displayADC);
    util::adc::start_async<*poti1Pin>();
}

ISR(TIMER1_OVF_vect) { startReadADC(); }

int main() {
    // configure pins
    pin::configure<*ledPin>     (OUTPUT);
    pin::configure<*internalLed>(OUTPUT);
    pin::configure<*btnPin>     (INPUT );
    pin::configure<*poti1Pin>   (INPUT );
    pin::configure<*poti2Pin>   (INPUT );
    pin::configure<*blPin>      (OUTPUT);

    // use timer2 for pwm
    util::timer::use_pwm<device::timer2, *blPin>();

    // use timer1 to read adc
    using t1 = device::timer1;
    t1::set_clock<t1::clock::ps8>();
    t1::set_waveform<t1::waveform::normal>();
    device::irq::enable<device::irq::timer1_ovf>();

    // hook up pushbutton
    util::irq::use_external<device::irq::int0>(buttonPressed);

    // switch on modules
    util::adc::use();
    device::irq::enable<device::irq::adc>();

    util::uart::use();
    
    lcd.clear();
    signalStartup();
    device::irq::enable_global();

    while(true) {}
}
